#!/usr/bin/env python
# encoding: utf-8
"""
copyright (c) 2024- Earth Advantage.
All rights reserved
..codeauthor::Fable Turas <fable@rainsoftware.tech>

"""

# Imports from Standard Library
import os
import secrets
import tempfile
from collections import OrderedDict
from unittest import TestCase, mock
from urllib.parse import urlparse, parse_qsl

# Imports from Third Party Modules
from jwcrypto import jwk
from jwcrypto.common import json_decode

# Imports from Django

# Local Imports
from logingov.tests.mock_server import MockOIDCServer, BASE_URL, DISCOVERY
from logingov import oidc
# Setup

# Constants

# Data Structure Definitions


# Private Functions
def discovery(self):
    return DISCOVERY


# Public Classes and Functions
class OIDCTests(TestCase):

    def setUp(self):
        self.client_id = secrets.token_urlsafe(16)

        pvt_key_path = os.path.join(tempfile.gettempdir(), 'pvt_key')
        self.redirect_uri = 'http://localhost:8000/oidc/callback'
        self.settings = {
            'CLIENT_ID': self.client_id,
            'LOGIN_REDIRECT_URI': self.redirect_uri,
            'LOGOUT_REDIRECT_URI': 'http://localhost:8000/oidc/logout',
            'PVT_KEY_PATH': pvt_key_path,
            'BASE_URL': BASE_URL,
        }
        pkjwk = jwk.JWK.generate(kty='RSA', size=4096)
        pk_attrs = json_decode(pkjwk.export_public())
        self.pk_pem = pkjwk.export_to_pem(True, None).decode('utf-8')
        with open(pvt_key_path, 'w') as file:
            file.write(self.pk_pem)
        pub_jwk = jwk.JWK()
        pub_jwk.import_key(**pk_attrs)
        pub_pem = pub_jwk.export_to_pem().decode('utf-8')
        self.server = MockOIDCServer(
            self.client_id, pub_pem, self.redirect_uri
        )

    @mock.patch.object(oidc.LoginGovOIDCClient, '_discovery', discovery)
    def test__get_base_url(self):
        settings = dict(self.settings)
        settings['BASE_URL'] = ''
        settings['ENVIRONMENT'] = 'sandbox'
        client = oidc.LoginGovOIDCClient(settings=settings)
        self.assertEqual(client._get_base_url(), oidc.SANDBOX_URL)

        settings['ENVIRONMENT'] = 'production'
        client = oidc.LoginGovOIDCClient(settings=settings)
        self.assertEqual(client._get_base_url(), oidc.PRODUCTION_URL)

    @mock.patch.object(oidc.LoginGovOIDCClient, '_discovery', discovery)
    def test__validate_settings(self):
        client = oidc.LoginGovOIDCClient(settings=self.settings)
        self.assertIsNone(client._validate_settings())

        settings = dict(self.settings)
        settings['SCOPE'] = 'openid email occupation'
        with self.assertRaises(oidc.LoginGovOIDCConfigurationError) as err:
            _ = oidc.LoginGovOIDCClient(settings=settings)
            self.assertIn('SCOPE', str(err))

        settings = dict(self.settings)
        settings['SIGNING_ALGORITHM'] = 'invalid_value'
        with self.assertRaises(oidc.LoginGovOIDCConfigurationError) as err:
            _ = oidc.LoginGovOIDCClient(settings=settings)
            self.assertIn('SIGNING_ALGORITHM', str(err))

    def test__discovery(self):
        with mock.patch('logingov.oidc.requests.get',
                        side_effect=self.server.route_request):
            client = oidc.LoginGovOIDCClient(settings=self.settings)
            self.assertDictEqual(client.discovery_data, DISCOVERY)

    @mock.patch.object(oidc.LoginGovOIDCClient, '_discovery', discovery)
    def test_authorize_redirect(self):
        client = oidc.LoginGovOIDCClient(settings=self.settings)
        with self.assertRaises(oidc.LoginGovOIDCConfigurationError):
            _ = client.authorize_redirect

        state = secrets.token_urlsafe(16)
        nonce = secrets.token_urlsafe(16)
        client = oidc.LoginGovOIDCClient(
            state=state, nonce=nonce, settings=self.settings,
            locale='es'
        )
        result = client.authorize_redirect
        self.assertIn(f'state={state}', result)
        self.assertIn(f'nonce={nonce}', result)
        self.assertIn(f'client_id={self.client_id}', result)
        self.assertIn('locale=es', result)

    @mock.patch.object(oidc.LoginGovOIDCClient, '_discovery', discovery)
    def test__validate_authorization_response(self):
        state = secrets.token_urlsafe(16)
        nonce = secrets.token_urlsafe(16)
        client = oidc.LoginGovOIDCClient(
            state=state, nonce=nonce, settings=self.settings
        )
        uri = client.authorize_redirect
        auth_response = self.server.authorize(uri)
        query = urlparse(auth_response).query
        auth_response = OrderedDict(parse_qsl(query))
        result = client._validate_authorization_response(auth_response)
        self.assertIsNone(result)  # no error raised

        # no code in response
        auth_response['code'] = None
        self.assertRaises(
            oidc.LoginGovOIDCAuthorizationError,
            client._validate_authorization_response, auth_response
        )

        # no state in client
        client = oidc.LoginGovOIDCClient(settings=self.settings)
        auth_response['code'] = secrets.token_urlsafe(16)
        self.assertRaises(
            oidc.LoginGovOIDCConfigurationError,
            client._validate_authorization_response, auth_response
        )

        # client state does not match response state
        client = oidc.LoginGovOIDCClient(
            state=secrets.token_urlsafe(16), settings=self.settings
        )
        self.assertRaises(
            oidc.LoginGovOIDCAuthorizationError,
            client._validate_authorization_response, auth_response
        )

        # unregistered client ID
        settings = dict(self.settings)
        settings['CLIENT_ID'] = secrets.token_urlsafe(16)
        client = oidc.LoginGovOIDCClient(
            state=state, nonce=nonce, settings=settings
        )
        uri = client.authorize_redirect
        auth_response = self.server.authorize(uri)
        query = urlparse(auth_response).query
        auth_response = OrderedDict(parse_qsl(query))
        self.assertRaises(
            oidc.LoginGovOIDCAuthorizationError,
            client._validate_authorization_response, auth_response
        )

    @mock.patch.object(oidc.LoginGovOIDCClient, '_discovery', discovery)
    def test_token_response(self):
        state = secrets.token_urlsafe(16)
        nonce = secrets.token_urlsafe(16)

        # missing authorization_response
        client = oidc.LoginGovOIDCClient(
            state=state, nonce=nonce, settings=self.settings
        )

        with self.assertRaises(oidc.LoginGovOIDCConfigurationError):
            _ = client.token_response

        # valid response
        uri = client.authorize_redirect
        auth_response = self.server.authorize(uri)
        client = oidc.LoginGovOIDCClient(
            state=state, nonce=nonce, settings=self.settings,
            authorization_response=auth_response
        )
        with mock.patch('logingov.oidc.requests.post',
                        side_effect=self.server.route_request):
            with mock.patch('logingov.oidc.requests.get',
                            side_effect=self.server.route_request):
                result = client.token_response
                self.assertIn('id_token', result)

        # valid response with pem instead of path
        settings = dict(self.settings)
        settings['PVT_KEY_PATH'] = ''
        settings['PVT_KEY_PEM'] = self.pk_pem
        client = oidc.LoginGovOIDCClient(
            state=state, nonce=nonce, settings=settings,
            authorization_response=auth_response
        )
        with mock.patch('logingov.oidc.requests.post',
                        side_effect=self.server.route_request):
            with mock.patch('logingov.oidc.requests.get',
                            side_effect=self.server.route_request):
                result = client.token_response
                self.assertIn('id_token', result)

        # invalid id_token (invalid aud)
        settings = dict(self.settings)
        settings['CLIENT_ID'] = secrets.token_urlsafe(16)
        client = oidc.LoginGovOIDCClient(
            state=state, nonce=nonce,
            settings=settings,
            authorization_response=auth_response
        )
        with mock.patch('logingov.oidc.requests.post',
                        side_effect=self.server.route_request):
            with mock.patch('logingov.oidc.requests.get',
                            side_effect=self.server.route_request):
                with self.assertRaises(oidc.LoginGovOIDCInvalidTokenError):
                    _ = client.token_response

        # no nonce
        client = oidc.LoginGovOIDCClient(
            state=state, settings=self.settings,
            authorization_response=auth_response
        )
        with mock.patch('logingov.oidc.requests.post',
                        side_effect=self.server.route_request):
            with mock.patch('logingov.oidc.requests.get',
                            side_effect=self.server.route_request):
                with self.assertRaises(oidc.LoginGovOIDCConfigurationError):
                    _ = client.token_response

        # invalid nonce
        client = oidc.LoginGovOIDCClient(
            state=state, nonce=secrets.token_urlsafe(16),
            settings=self.settings,
            authorization_response=auth_response
        )
        with mock.patch('logingov.oidc.requests.post',
                        side_effect=self.server.route_request):
            with mock.patch('logingov.oidc.requests.get',
                            side_effect=self.server.route_request):
                with self.assertRaises(oidc.LoginGovOIDCInvalidTokenError):
                    _ = client.token_response

        # invalid code
        query = urlparse(auth_response).query
        auth_response = OrderedDict(parse_qsl(query))
        auth_response['code'] = secrets.token_urlsafe(16)
        client = oidc.LoginGovOIDCClient(
            state=state, nonce=nonce, settings=self.settings,
            authorization_response=auth_response
        )
        with mock.patch('logingov.oidc.requests.post',
                        side_effect=self.server.route_request):
            with mock.patch('logingov.oidc.requests.get',
                            side_effect=self.server.route_request):
                with self.assertRaises(oidc.LoginGovOIDCRequestError):
                    _ = client.token_response

    @mock.patch.object(oidc.LoginGovOIDCClient, '_discovery', discovery)
    def test_userinfo_id_token(self):
        state = secrets.token_urlsafe(16)
        nonce = secrets.token_urlsafe(16)
        client = oidc.LoginGovOIDCClient(
            state=state, nonce=nonce, settings=self.settings
        )
        uri = client.authorize_redirect
        auth_response = self.server.authorize(uri)
        client = oidc.LoginGovOIDCClient(
            state=state, nonce=nonce, settings=self.settings,
            authorization_response=auth_response
        )
        with mock.patch('logingov.oidc.requests.post',
                        side_effect=self.server.route_request):
            with mock.patch('logingov.oidc.requests.get',
                            side_effect=self.server.route_request):
                result = client.userinfo_id_token
                self.assertIn('email', result[0])

    @mock.patch.object(oidc.LoginGovOIDCClient, '_discovery', discovery)
    def test_logout_redirect(self):
        state = secrets.token_urlsafe(16)
        client = oidc.LoginGovOIDCClient(
            state=state, settings=self.settings
        )
        result = client.logout_redirect
        self.assertIn(f'client_id={self.client_id}', result)

        settings = dict(self.settings)
        settings['LOGOUT_REDIRECT_URI'] = ''

        client = oidc.LoginGovOIDCClient(
            state=state, settings=settings
        )
        with self.assertRaises(oidc.LoginGovOIDCConfigurationError):
            _ = client.logout_redirect



#!/usr/bin/env python
# encoding: utf-8
"""
copyright (c) 2024- Earth Advantage.
All rights reserved
..codeauthor::Fable Turas <fable@rainsoftware.tech>

"""

# Imports from Standard Library
import secrets
import time
from collections import OrderedDict
from urllib.parse import urljoin, urlencode, urlparse, parse_qsl

# Imports from Third Party Modules
import jwt
import requests
from jwcrypto import jwk

# Local Imports
from logingov.settings.oidc_settings import (
    oidc_settings,
    OIDCSettings,
    DEFAULTS
)
from logingov.helpers import load_private_key, encode_left128bits

# Setup

# Constants
SANDBOX_URL = 'https://idp.int.identitysandbox.gov/'
PRODUCTION_URL = 'https://secure.login.gov/'
CONFIG_SETTINGS_ATTRIBS = {
    'ACR_VALUES': 'acr_values_supported',
    'SIGNING_ALGORITHM': 'token_endpoint_auth_signing_alg_values_supported',
    'GRANT_TYPE': 'grant_types_supported',
    'RESPONSE_TYPE': 'response_types_supported',
    'SCOPE': 'scopes_supported',
}
SPACE_SEPARATED_ATTRIBS = [
    'ACR_VALUES',
    'SCOPE'
]

# Data Structure Definitions

# Private Functions


# Public Classes and Functions
class LoginGovOIDCClientError(Exception):
    pass


class LoginGovOIDCAuthorizationError(LoginGovOIDCClientError):
    """Exception for errors in the Login.gov authorization response"""
    pass


class LoginGovOIDCRequestError(LoginGovOIDCClientError):
    """Exception for raising errors from Login.gov API requests"""
    pass


class LoginGovOIDCInvalidTokenError(LoginGovOIDCClientError):
    """
    Exception for raising errors when Login.gov tokens are invalid
    or cannot be decoded
    """
    pass


class LoginGovOIDCConfigurationError(LoginGovOIDCClientError):
    """
    Exception for raising errors where the provided settings or
    client instantiation parameters are invalid
    """
    pass


class LoginGovOIDCClient(object):
    """
    A client for handling login.gov OIDC requests and responses.
    The client is framework-agnostic, and can be used in any Python
    application that needs to interact with the login.gov OIDC service.

    The client does not generate state or nonce values, or provide for
    storage of these values for use in later calls.  The calling code
    is responsible for generating and managing these values.

    Since the authorization is a redirect to the login.gov site, and the
    response is a redirect back to the client uri, with query params,
    the client does not handle the actual redirect. Therefore, it cannot
    directly call for the authorization response. The calling code must
    handle the redirect and pass the query params to the client, as a
    dict, or query string, for token and user info requests.

    Most of the client parameters are handled via settings. Settings must be
    passed to the client on instantiation as a dict or OIDCSettings instance,
    or may be configured within Django settings under the `LOGIN_GOV_OIDC`
    space (assuming the calling code is a Django app and has added `logingov`
    to the `INSTALLED_APPS`).
    """

    def __init__(self, state=None, nonce=None, locale=None,
                 authorization_response=None, settings=None):
        if isinstance(settings, OIDCSettings):               # pragma: no cover
            self.settings = settings
        else:
            self.settings = OIDCSettings(
                user_settings=settings, defaults=DEFAULTS
            ) if settings else oidc_settings
        self.settings.check_required_settings()

        self.state = state
        self.nonce = nonce
        self.locale = locale
        self.discovery_data = self._discovery()
        self._validate_settings()
        self.code = None
        self.jti = None
        if authorization_response:
            if isinstance(authorization_response, str):
                query = urlparse(authorization_response).query
                authorization_response = OrderedDict(parse_qsl(query))
            self._validate_authorization_response(authorization_response)
            self.code = authorization_response.get('code')

    def _get_base_url(self):
        """
        Get the base url for talking to the login.gov servers
        based on the `ENVIRONMENT` setting or the `BASE_URL` setting.
        """
        base_url = self.settings.BASE_URL
        env = self.settings.ENVIRONMENT
        if base_url:
            return base_url
        else:
            if env.lower().startswith('sand'):
                return SANDBOX_URL
            if env.lower().startswith('prod'):
                return PRODUCTION_URL

    def _discovery(self):
        """
        Uses the base url and DISCOVERY_ENDPOINT to get the OIDC discovery
        data. All other endpoints are pulled from the discovery results,
        rather than being hard-coded, or settings values, to ensure the client
        is always using the most up-to-date login.gov information.
        """
        endpoint = self.settings.DISCOVERY_ENDPOINT
        url = urljoin(self._get_base_url(), endpoint)
        return self.get(url)

    def _validate_settings(self):
        """
        Validates the relevant settings values against the discovery data
        supported values.
        """
        invalid = []
        for setting, config_key in CONFIG_SETTINGS_ATTRIBS.items():
            setting_val = getattr(self.settings, setting)
            config_val = self.discovery_data[config_key]
            if setting in SPACE_SEPARATED_ATTRIBS:
                setting_val = setting_val.split(' ')
                if any([val not in config_val for val in setting_val]):
                    invalid.append(setting)
            elif setting_val not in config_val:
                invalid.append(setting)
        if invalid:
            raise LoginGovOIDCConfigurationError(
                f'The following "LOGIN_GOV_OIDC" settings have values '
                f'that do not match the Login.gov supported values: '
                f'{", ".join(invalid)}'
            )

    def _prepare_authorize_params(self):
        """
        Prepares the query string parameters for inclusion in the authorization
        redirect URL based on provided settings and the client instantiation
        values.
        Both state and nonce are required, and should be setup and managed
        inside the code calling the client.  The same state and nonce will be
        needed again later to validate the authorization and token responses.
        """
        if not self.state or not self.nonce:
            raise LoginGovOIDCConfigurationError(
                'You must provide a state and nonce to the client '
                'to prepare an authorization redirect'
            )
        scope = self.settings.SCOPE.split(' ')
        if 'openid' not in scope:
            scope.insert(0, 'openid')
        params = {
            'acr_values': self.settings.ACR_VALUES,
            'client_id': self.settings.CLIENT_ID,
            'prompt': self.settings.PROMPT,
            'response_type': self.settings.RESPONSE_TYPE,
            'redirect_uri': self.settings.LOGIN_REDIRECT_URI,
            'state': self.state,
            'nonce': self.nonce,
        }
        if self.locale in self.settings.ACCEPTED_LOCALES:
            params['locale'] = self.locale
        return f'?{urlencode(params)}&scope={"+".join(scope)}'

    def _validate_authorization_response(self, auth_response):
        """
        Ensures the authorization response is not an error, contains a code,
        and that the state used to generate the authorization request matches
        the state returned in the response.
        """
        err = auth_response.get('error') or ''
        if err:
            err_desc = auth_response.get('error_description') or ''
            err_msg = ': '.join([err, err_desc])
            raise LoginGovOIDCAuthorizationError(err_msg)
        if not auth_response.get('code'):
            raise LoginGovOIDCAuthorizationError(
                'No authorization code provided'
            )
        if not self.state:
            raise LoginGovOIDCConfigurationError(
                'You must provide the state from your authorization request '
                'to the client to validate the authorization response.'
            )
        if auth_response.get('state') != self.state:
            raise LoginGovOIDCAuthorizationError('Invalid state in response')

    def _prepare_token_assertion(self):
        """Prepares the client assertion JWT for use in the token request."""
        self.jti = secrets.token_hex(16)
        args = {
            'iss': self.settings.CLIENT_ID,
            'sub': self.settings.CLIENT_ID,
            'aud': self.discovery_data['token_endpoint'],
            'jti': self.jti,
            'exp': int(time.time()) + 300,
        }
        if self.settings.PVT_KEY_PATH:
            key = load_private_key(self.settings.PVT_KEY_PATH)
        else:
            key = self.settings.PVT_KEY_PEM
        return jwt.encode(args, key, algorithm=self.settings.SIGNING_ALGORITHM)

    def _prepare_token_params(self):
        """Prepares the POST data for use in the token request."""
        return {
            'client_assertion': self._prepare_token_assertion(),
            'client_assertion_type': self.settings.CLIENT_ASSERTION_TYPE,
            'code': self.code,
            'grant_type': self.settings.GRANT_TYPE,
        }

    def _login_gov_pub_keys(self):
        """
        Gets the login.gov public keys for use in decoding and validating
        the token request response.
        """
        return self.get(self.discovery_data['jwks_uri'])['keys']

    def _decoded_id_token(self, id_token):
        """
        Decodes the ID token using the login.gov public keys.
        A LoginGovOIDCInvalidTokenError will be raised if the ID token cannot
        be decoded, or the aud, iss, exp, nbf, or iat values are determined
        to be invalid during the decoding.
        """
        for key in self._login_gov_pub_keys():
            jwk_key = jwk.JWK(**key)
            pem = jwk_key.export_to_pem().decode('utf-8')
            try:
                return jwt.decode(
                    id_token, pem,
                    algorithms=[self.settings.SIGNING_ALGORITHM],
                    audience=[self.settings.CLIENT_ID],
                    issuer=self.discovery_data['issuer'],
                    leeway=2
                )
            except Exception:
                continue
        else:
            raise LoginGovOIDCInvalidTokenError('Unable to decode ID token')

    def _validated_token(self, tokens):
        """
        Ensures the validity of the token response.
        Raises a LoginGovOIDCInvalidTokenError if the nonce, access token hash,
        code hash, or any of the base JWT values are invalid.
        """
        if not self.nonce:
            raise LoginGovOIDCConfigurationError(
                'You must provide the nonce from your authorization request '
                'to the client to validate the authorization token.'
            )
        id_token = self._decoded_id_token(tokens['id_token'])
        if id_token['nonce'] != self.nonce:
            raise LoginGovOIDCInvalidTokenError('Invalid nonce in ID token')
        if id_token['at_hash'] != encode_left128bits(tokens['access_token']):  # pragma: no cover  # noqa
            raise LoginGovOIDCInvalidTokenError(
                'Invalid access token hash in ID token'
            )
        if id_token['c_hash'] != encode_left128bits(self.code):  # pragma: no cover  # noqa
            raise LoginGovOIDCInvalidTokenError(
                'Invalid code hash in ID token'
            )
        tokens['id_token'] = id_token
        return tokens

    @staticmethod
    def post(url, data):
        """
        Sends a POST request to the provided URL with the provided data,
        handling any error statuses as a LoginGovOIDCRequestError.
        """
        rsp = requests.post(url, data=data)
        if rsp.status_code != 200:
            raise LoginGovOIDCRequestError(
                f'Error {rsp.status_code} from {url}: {rsp.json()["error"]}'
            )
        return rsp.json()

    @staticmethod
    def get(url, params=None, headers=None):
        """
        Sends a GET request to the provided URL with the provided params and/or
        headers, handling any error statuses as a LoginGovOIDCRequestError.
        """
        rsp = requests.get(url, params=params, headers=headers)
        if rsp.status_code != 200:
            raise LoginGovOIDCRequestError(
                f'Error {rsp.status_code} from {url}: {rsp.json()["error"]}'
            )
        return rsp.json()

    @property
    def authorize_redirect(self):
        """Generate the login.gov authorization redirect URI."""
        url = self.discovery_data['authorization_endpoint']
        return urljoin(url, self._prepare_authorize_params())

    @property
    def token_response(self):
        """
        Request tokens from login.gov using the code from
        the authorization login redirect.
        """
        if not self.code:
            raise LoginGovOIDCConfigurationError(
                'You must complete the Login.gov authorization and pass '
                'the results to the client before requesting tokens '
                'or user info.'
            )
        url = self.discovery_data['token_endpoint']
        token_rsp = self.post(url, data=self._prepare_token_params())
        return self._validated_token(token_rsp)

    @property
    def userinfo_id_token(self):
        """
        Request tokens and user info from login.gov using the code from
        the authorization login redirect.
        """
        token_rsp = self.token_response
        headers = {
            'Authorization':
                f'{token_rsp["token_type"]} {token_rsp["access_token"]}'
        }
        url = self.discovery_data['userinfo_endpoint']
        return self.get(url, headers=headers), token_rsp['id_token']

    @property
    def logout_redirect(self):
        """Generate the login.gov logout redirect URI."""
        redirect = self.settings.LOGOUT_REDIRECT_URI
        if not redirect:
            raise LoginGovOIDCConfigurationError(
                'You must configure a LOGOUT_REDIRECT_URI.'
            )
        params = {
            'client_id': self.settings.CLIENT_ID,
            'post_logout_redirect_uri': redirect
        }
        if self.state:
            params['state'] = self.state
        url = self.discovery_data['end_session_endpoint']
        return urljoin(url, f'?{urlencode(params)}')



#!/usr/bin/env python
# encoding: utf-8
"""
copyright (c) 2024- Earth Advantage.
All rights reserved
..codeauthor::Fable Turas <fable@rainsoftware.tech>

Settings for py-logingov-client are all namespaced in the
LOGIN_GOV_OIDC setting.
For example your project's `settings.py` file might look like this:

LOGIN_GOV_OIDC = {
}

This module provides the `oidc_setting` object, that is used to access
py-logingov settings, checking for user settings first, then falling
back to the defaults."""

# Imports from Standard Library

# Imports from Third Party Modules

# Imports from Django
try:                                                         # pragma: no cover
    from django.conf import settings
    from django.core.signals import setting_changed
except ImportError:
    settings = None

# Local Imports

# Setup

# Constants
DEFAULTS = {
    'CLIENT_ID': '',
    'LOGIN_REDIRECT_URI': '',
    'LOGOUT_REDIRECT_URI': '',
    'PVT_KEY_PATH': '',
    'PVT_KEY_PEM': '',
    'ACR_VALUES': 'http://idmanagement.gov/ns/assurance/ial/1',
    'SCOPE': 'openid email',
    'PROMPT': 'select_account',
    'RESPONSE_TYPE': 'code',
    'GRANT_TYPE': 'authorization_code',
    'CLIENT_ASSERTION_TYPE': (
        'urn:ietf:params:oauth:client-assertion-type:jwt-bearer'
    ),
    'SIGNING_ALGORITHM': 'RS256',
    'DISCOVERY_ENDPOINT': '.well-known/openid-configuration',
    'ENVIRONMENT': 'sandbox',
    'BASE_URL': '',
    'ACCEPTED_LOCALES': ['es', 'fr']
}
REQUIRED_SETTINGS = (
    'CLIENT_ID',
    'LOGIN_REDIRECT_URI',
    'SCOPE',
    'ACR_VALUES'
)
ALTERNATE_REQUIRED_SETTINGS = (
    ('PVT_KEY_PATH', 'PVT_KEY_PEM'),
    ('ENVIRONMENT', 'BASE_URL')
)
# Data Structure Definitions

# Private Functions


# Public Classes and Functions
class OIDCSettings:
    """
    A settings object, that allows settings to be accessed as properties.
    For example:

        from django_logingov_oidc.settings import oidc_settings
        print(oidc_settings.DEFAULTS)

    For Django usage, settings must be namespaced under
    the `LOGIN_GOV_OIDC` name.
    """

    def __init__(self, user_settings=None, defaults=None):
        self._user_settings = user_settings

        self.defaults = defaults or DEFAULTS
        self._registered_attrs = set()

    @property
    def user_settings(self):
        return (
            self._user_settings or getattr(settings, "LOGIN_GOV_OIDC", {})
        )

    def __getattr__(self, attr):
        if attr != 'user_settings' and attr not in self.defaults:
            raise AttributeError(f"Invalid OIDC setting: '{attr}'")
        try:
            val = self.user_settings[attr]
        except KeyError:
            val = self.defaults[attr]

        self._registered_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def check_required_settings(self):
        """Validates that all required settings are present and not empty"""
        for setting in REQUIRED_SETTINGS:
            setting_val = getattr(self, setting, None)
            if setting_val in ['', None]:
                raise ValueError(f"Setting {setting} must not be empty")
        for keys in ALTERNATE_REQUIRED_SETTINGS:
            vals = [
                getattr(self, setting, None)
                for setting in keys
            ]
            if not any(vals):
                raise ValueError(
                    f'At least one of {", ".join(keys)} must not be empty'
                )

    def reload(self):
        """
        Resets all registered settings to ensure prior set values are not
        lingering when Django settings are changed.
        """
        for attr in self._registered_attrs:
            delattr(self, attr)
        self._registered_attrs.clear()
        self._user_settings = {}


oidc_settings = OIDCSettings(None, DEFAULTS)


def reload_oidc_settings(setting=None, *args, **kwargs):     # pragma: no cover
    if setting == "LOGIN_GOV_OIDC":
        oidc_settings.reload()


if settings:                                                 # pragma: no cover
    setting_changed.connect(reload_oidc_settings)



#!/usr/bin/env python
# encoding: utf-8
"""
copyright (c) 2024- Earth Advantage.
All rights reserved
..codeauthor::Fable Turas <fable@rainsoftware.tech>

"""

# Imports from Standard Library
import secrets
import time
from collections import OrderedDict
from urllib.parse import urlparse, parse_qsl, urljoin, urlencode

# Imports from Third Party Modules
import jwt
from jwcrypto import jwk
from jwcrypto.common import json_decode

# Imports from Django

# Local Imports
from logingov.helpers import encode_left128bits

# Setup

# Constants
BASE_URL = 'http://testserver'

DISCOVERY = {'acr_values_supported': [
    'http://idmanagement.gov/ns/assurance/loa/1',
    'http://idmanagement.gov/ns/assurance/loa/3',
    'http://idmanagement.gov/ns/assurance/ial/1',
    'http://idmanagement.gov/ns/assurance/ial/2',
    'http://idmanagement.gov/ns/assurance/ial/0',
    'http://idmanagement.gov/ns/assurance/ial/2?strict=true',
    'urn:gov:gsa:ac:classes:sp:PasswordProtectedTransport:duo',
    'http://idmanagement.gov/ns/assurance/aal/2',
    'http://idmanagement.gov/ns/assurance/aal/3',
    'http://idmanagement.gov/ns/assurance/aal/3?hspd12=true',
    'http://idmanagement.gov/ns/assurance/aal/2?phishing_resistant=true',
    'http://idmanagement.gov/ns/assurance/aal/2?hspd12=true'
],
    'claims_supported': [
        'iss',
        'sub',
        'email',
        'email_verified',
        'all_emails',
        'address',
        'phone',
        'phone_verified',
        'given_name',
        'family_name',
        'birthdate',
        'verified_at',
        'social_security_number',
        'x509_subject',
        'x509_presented',
        'x509_issuer'
    ],
    'grant_types_supported': ['authorization_code'],
    'response_types_supported': ['code'],
    'scopes_supported': [
        'email',
        'all_emails',
        'openid',
        'profile:verified_at',
        'x509',
        'x509:subject',
        'x509:issuer',
        'x509:presented',
        'address',
        'phone',
        'profile',
        'profile:name',
        'profile:birthdate',
        'social_security_number'
    ],
    'subject_types_supported': ['pairwise'],
    'authorization_endpoint': f'{BASE_URL}/openid_connect/authorize',
    'issuer': BASE_URL,
    'jwks_uri': f'{BASE_URL}/api/openid_connect/certs',
    'service_documentation': 'https://developers.login.gov/',
    'token_endpoint': f'{BASE_URL}/api/openid_connect/token',
    'userinfo_endpoint': f'{BASE_URL}/api/openid_connect/userinfo',
    'end_session_endpoint': f'{BASE_URL}/openid_connect/logout',
    'id_token_signing_alg_values_supported': ['RS256'],
    'token_endpoint_auth_methods_supported': ['private_key_jwt'],
    'token_endpoint_auth_signing_alg_values_supported': ['RS256']
}

USER_PAYLOAD = {
    'address': {
        'formatted': '123 Main St Apt 123\nWashington, DC 20001',
        'street_address': '123 Main St Apt 123',
        'locality': 'Washington',
        'region': 'DC',
        'postal_code': '20001'
    },
    'birthdate': '1970-01-01',
    'email': 'test@example.com',
    'email_verified': True,
    'all_emails': ['test@example.com', 'test2@example.com'],
    'family_name': 'Smith',
    'given_name': 'John',
    'phone': '+18881112222',
    'phone_verified': True,
    'social_security_number': '111223333',
    'sub': 'test-uuid',
    'verified_at': 1577854800
}
SCOPE = {
    'email': ['email', 'email_verified'],
    'profile': ['family_name', 'given_name', 'birthdate'],
    'profile:name': ['family_name', 'given_name'],
    'profile:verified_at': ['verified_at'],
    'address': ['address'],
    'phone': ['phone'],
    'social_security_number': ['social_security_number']
}
# Data Structure Definitions

# Private Functions


# Public Classes and Functions
class MockResponse:
    def __init__(self, json_rsp, status_code):
        self.json_rsp = json_rsp
        self.status_code = status_code

    def json(self):
        return self.json_rsp


class MockOIDCServer(object):
    authorized = {}

    def __init__(self, client_id, public_key, redirect_uri, algo=None):
        self.jwk = jwk.JWK.generate(kty='RSA', size=4096)
        self.pk_attrs = json_decode(self.jwk.export_public())
        self.pk_pem = self.jwk.export_to_pem(True, None).decode('utf-8')
        self.pub_jwk = jwk.JWK()
        self.pub_jwk.import_key(**self.pk_attrs)
        self.pub_key = self.pub_jwk.export_public(True)
        self.client_registry = {
            client_id: {
                'key': public_key,
                'redirect_uri': redirect_uri,
            }
        }
        self.signing_algo = algo if algo else 'RS256'

    @staticmethod
    def _nested_find(data, nested_key, nested_val):
        for top_key, sub_dict in data.items():
            lookup = sub_dict.get(nested_key)
            if lookup == nested_val:
                return data[top_key]
        return None

    @property
    def routes(self):
        return {
            f'{BASE_URL}/.well-known/openid-configuration': self.discovery,
            DISCOVERY['authorization_endpoint']: self.authorize,
            DISCOVERY['jwks_uri']: self.keys,
            DISCOVERY['token_endpoint']: self.token,
            DISCOVERY['userinfo_endpoint']: self.userinfo,
        }

    def route_request(self, uri, data=None, headers=None, params=None):
        route = uri.split('?')[0]
        route_method = self.routes[route]
        return route_method(data=data, headers=headers, params=params)

    def discovery(self, *args, **kwargs):
        return MockResponse(DISCOVERY, 200)

    def authorize(self, uri, **kwargs):
        query = urlparse(uri).query
        query = OrderedDict(parse_qsl(query))
        client = self.client_registry.get(query['client_id'])
        if not client or query['redirect_uri'] != client['redirect_uri']:
            params = {
                'error': 'invalid_request ',
                'error_description': 'invalid request'
            }
        else:
            code = secrets.token_hex(16)
            scope = query['scope'].replace('+', ' ').split(' ')
            self.authorized[code] = {
                'client_id': query['client_id'],
                'state': query['state'],
                'nonce': query['nonce'],
                'acr': query['acr_values'],
                'scope': scope,
                'access_token': secrets.token_hex(16),
            }
            params = {'code': code, 'state': query['state']}
        uri = urljoin(query['redirect_uri'], f'?{urlencode(params)}')
        return uri

    def keys(self, *args, **kwargs):
        return MockResponse(
            {
                'keys': [
                    {
                        **self.pub_key,
                        'kid': self.pub_jwk.thumbprint(),
                    }
                ]
            },
            200
        )

    def token(self, data=None, *args, **kwargs):
        code = data['code']
        auth = self.authorized.get(code)
        if not auth:
            return MockResponse(
                {'error': 'invalid code'},
                400
            )
        client = self.client_registry[auth['client_id']]
        client_assertion = data['client_assertion']
        client_jwt = jwt.decode(
            client_assertion, client['key'],
            algorithms=[self.signing_algo],
            options={'verify_signature': False}
        )
        id_params = {
            'iss': BASE_URL,
            'sub': 'test-uuid',
            'aud': auth['client_id'],
            'acr': auth['acr'],
            'at_hash': encode_left128bits(auth['access_token']),
            'c_hash': encode_left128bits(code),
            'exp': time.time() + 60,
            'iat': time.time(),
            'jti': client_jwt['jti'],
            'nbf': time.time(),
            'nonce': auth['nonce']
        }
        id_token = jwt.encode(
            id_params, self.pk_pem, algorithm=self.signing_algo,
            headers={"kid": self.pub_jwk.thumbprint()},
        )
        rsp = {
            'id_token': id_token,
            'expires_in': 60,
            'token_type': 'Bearer',
            'access_token': auth['access_token']
        }
        return MockResponse(rsp, 200)

    def userinfo(self, headers=None, **kwargs):
        token = headers['Authorization'].split(' ')[1]
        auth = self._nested_find(self.authorized, 'access_token', token)
        if not auth:
            return MockResponse({'error': 'invalid token'}, 401)
        payload = {
            'iss': BASE_URL,
        }
        for scope, attrs in SCOPE.items():
            if scope in auth['scope']:
                payload.update({attr: USER_PAYLOAD[attr] for attr in attrs})
        return MockResponse(payload, 200)



#!/usr/bin/env python
# encoding: utf-8
"""
copyright (c) 2024- Earth Advantage.
All rights reserved
..codeauthor::Fable Turas <fable@rainsoftware.tech>

"""

# Imports from Standard Library
import base64
import hashlib

# Imports from Third Party Modules

# Imports from Django

# Local Imports

# Setup

# Constants

# Data Structure Definitions

# Private Functions


# Public Classes and Functions
def load_private_key(key_path):
    """load private key from file for encoding."""
    with open(key_path, 'rb') as pem_in:
        return pem_in.read()


def encode_left128bits(string):
    # 128 bits / 8 bits per byte = 16
    # ref https://github.com/18F/identity-idp/blob/799fc62621a30c54e7edba17e376d94606d0c956/app/services/id_token_builder.rb#L69  # noqa
    hashed = hashlib.sha256(string.encode("utf-8")).digest()[0:16]
    b64 = base64.urlsafe_b64encode(hashed)
    return b64.decode("utf-8").rstrip("=")



#!/usr/bin/env python
# encoding: utf-8
"""
copyright (c) 2024- Earth Advantage.
All rights reserved
..codeauthor::Fable Turas <fable@rainsoftware.tech>

"""

# Imports from Standard Library
import os
import secrets
import tempfile
from unittest import TestCase

# Imports from Django

# Local Imports
from logingov.tests.mock_server import BASE_URL
from logingov.settings.oidc_settings import OIDCSettings

# Setup

# Constants

# Data Structure Definitions

# Private Functions


# Public Classes and Functions
class OIDCTests(TestCase):

    def setUp(self):
        self.client_id = secrets.token_urlsafe(16)

        pvt_key_path = os.path.join(tempfile.gettempdir(), 'pvt_key')
        self.redirect_uri = 'http://localhost:8000/oidc/callback'
        self.user_settings = {
            'CLIENT_ID': self.client_id,
            'LOGIN_REDIRECT_URI': self.redirect_uri,
            'PVT_KEY_PATH': pvt_key_path,
            'BASE_URL': BASE_URL,
        }

    def test_check_required_settings(self):
        settings = OIDCSettings()
        self.assertRaises(
            ValueError,
            settings.check_required_settings
        )

        settings = OIDCSettings({
            'CLIENT_ID': self.client_id,
            'LOGIN_REDIRECT_URI': self.redirect_uri
        })
        self.assertRaises(
            ValueError,
            settings.check_required_settings
        )

        settings = OIDCSettings(user_settings=self.user_settings)
        self.assertIsNone(settings.check_required_settings())

    def test_reload(self):
        settings = OIDCSettings(user_settings=self.user_settings)
        self.assertDictEqual(settings.user_settings, self.user_settings)
        self.assertIsNotNone(settings.SCOPE)

        settings.reload()
        self.assertDictEqual(settings.user_settings, {})

    def test_getattr(self):
        settings = OIDCSettings(user_settings={'NOT_A_VALID_SETTING': 'test'})
        self.assertRaises(
            AttributeError,
            settings.__getattr__, 'NOT_A_VALID_SETTING'
        )



#!/usr/bin/env python
# encoding: utf-8
"""
copyright (c) 2023- Earth Advantage.
All rights reserved
..codeauthor::Fable Turas <fable@rainsoftware.tech>

"""

# Imports from Standard Library
import os

# Imports from Django
from django.core.management.base import BaseCommand

from logingov.settings.oidc_settings import oidc_settings

# Setup

# Constants

# Data Structure Definitions

# Private Functions


# Public Classes and Functions
class Command(BaseCommand):
    help = (
        'Write Private Key from env vars, or argument, to file'
    )

    def add_arguments(self, parser):                         # pragma: no cover
        parser.add_argument(
            '-k', '--key',
            help=(
                'Private key string, with escaped line '
                'breaks, to write to file.'
            )
        )
        parser.add_argument(
            '-e', '--env',
            help=(
                'Environment variable name where private key is stored'
            )
        )

    def handle(self, *args, **options):                      # pragma: no cover
        pk = options['key']
        env = options['env'] or 'LOGIN_GOV_OIDC_PVT_KEY'
        pk = pk if pk else os.environ[env]
        pk = pk.replace('\\n', '\n')
        with open(oidc_settings.PVT_KEY_PATH, 'w') as file:
            file.write(pk)
            self.stdout.write(self.style.SUCCESS(
                f'Successfully generated private key file '
                f'at {oidc_settings.PVT_KEY_PATH}'
            ))
